"""NLP components to be used within a spacy pipeline.

This module contains components (along with their helper functions) to be used
within spacy NLP pipelines. Each component takes a Doc as an input, assigns some
attribute to one of the core spacy object types (Doc, Span, or Token), and
returns a Doc containing the set attributes. These components can then be
assembled into a pipeline, building upon one another and customizing the spacy
default pipeline.
"""

from spacy.tokens import Doc, Span, Token
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer

# Initializing the VADER sentiment analyzer for use within sentiment components
ANALYZER = SentimentIntensityAnalyzer()

# Helper functions
def set_extension(
    extension_name: str,
    default_val: any = None,
    target_obj: Doc | Span | Token = Doc,
) -> None:
    """Sets extension on designated spacy object if it doesn't already exist.

    Args:
        extension_name (str): Name of the extension/attribute to be set.
        default_val (any, optional): Default value for all objects with
            the extension to be initialized to. Defaults to None.
        target_obj (Doc | Span | Token, optional): The spacy object onto which
            the extension should be set. Defaults to Doc.
    """
    if not target_obj.has_extension(extension_name):
        target_obj.set_extension(extension_name, default=default_val)


def get_token_parent_span(
    token: Token,
    min_length: int,
) -> Span:
    """Takes a Token object and returns its Span from the parent Doc.

    Args:
        token (Token): spacy Token object to get the parent span of.
        min_length (int): Minimum length of parent span. This will be ignored
            if the sentence itself is shorter min_length.

    Returns:
        Span: spacy Span object containing the Token passed into the function.
    """

    def get_children_indices(head: Token) -> list:
        """Recursively gets indices of head's children and children of children.

        Args:
            head (Token): Token to get the indices of children for.

        Returns:
            list: List of indices of child tokens (indices are the locations of
                the children within the parent Doc object).
        """
        indices = [token.i for token in head.children]
        indices.append(head.i)
        for child in head.children:
            indices.extend(get_children_indices(child))
        indices.sort()
        return indices

    indices = get_children_indices(token.head)
    doc = token.doc
    first = indices[0]
    last = indices[-1]
    if last >= len(doc):
        span = doc[first:]
    else:
        span = doc[first : last + 1]
    if len(span) < min_length and token.head != token.head.head:
        span = get_token_parent_span(token.head, min_length=min_length)
    return span


# Component functions
def set_doc_contains_aspect(
    doc: Doc,
    base_keywords: list,
) -> Doc:
    """Takes a Doc and returns a new Doc with the 'contains_aspect' attribute.

    Accessed via 'Doc._.contains_aspect', the 'contains_aspect' attribute is a
    boolean which indicates whether or not doc contains any of the keywords
    passed to function.

    Target object: spacy Doc
    Attribute type: bool
    Default value: False
    Dependency path: N/A

    Args:
        doc (Doc): The Doc object to set the attribute on.
        base_keywords (list): List of keywords to search for within the Doc.

    Returns:
        Doc: Processed Doc object with the 'contains_aspect' attribute.
    """
    set_extension("contains_aspect", default_val=False)

    for token in doc:
        if token.lemma_.lower() in base_keywords:
            doc._.contains_aspect = True
            break

    return doc


def set_doc_aspects(
    doc: Doc,
    base_aspects: dict,
) -> Doc:
    """Takes a Doc and returns a new Doc with the 'aspects' attribute.

    Accessed via 'Doc._.aspects', the 'aspects' attribute contains a list of the
    aspects discussed by the Doc.

    Target object: spacy Doc
    Attribute type: list
    Default value: None
    Dependency path: set_doc_contains_aspect() ->

    Args:
        doc (Doc): The Doc object to set the attribute on.
        base_aspects (dict): Dictionary of keywords mapped to aspects. Should
            take the form of: {'aspect1': ['keyword1', 'keyword2'], 'aspect2':
            ['keyword3', 'keyword4']}.

    Returns:
        Doc: Processed Doc object with the 'aspects' attribute.
    """
    set_extension("aspects")

    if doc._.contains_aspect == True:
        aspects_contained = []
        for token in doc:
            for aspect in base_aspects:
                if aspect in aspects_contained:
                    continue
                keywords = base_aspects[aspect]
                if token.lemma_.lower() in keywords:
                    aspects_contained.append(aspect)
        doc._.aspects = aspects_contained
    return doc


def set_doc_keywords(
    doc: Doc,
    base_keywords: list,
) -> Doc:
    """Takes a Doc and returns a new Doc with the 'keywords' attribute.

    Accessed via 'Doc._.keywords', the 'keywords' attribute contains a list of
    the keywords contained within the Doc.

    Target object: spacy Doc
    Attribute type: list
    Default value: None
    Dependency path: set_doc_contains_aspect() ->

    Args:
        doc (Doc): The Doc object to set the attribute on.
        base_keywords (list): List of keywords to search for within the Doc.

    Returns:
        Doc: Processed Doc object with the 'keywords' attribute.
    """
    set_extension("keywords")

    if doc._.contains_aspect == True:
        keywords = []
        for token in doc:
            if token.lemma_.lower() in base_keywords:
                keywords.append(token)
        doc._.keywords = keywords
    return doc


def set_token_aspects(
    doc: Doc,
    base_aspects: dict,
) -> Doc:
    """Takes a Doc and adds the 'aspect' attribute to its Token objects.

    Accessed via 'Token._.aspect', the 'aspect' attribute is applied only to the
    Token objects contained within the 'keywords' attribute of the Doc. The
    attribute itself reflects the corresponding aspect of the keyword. Non-
    keyword Token objects receive a None value.

    Target object: spacy Token
    Attribute type: string
    Default value: None
    Dependency path: set_doc_contains_aspect() -> set_doc_keywords() ->

    Args:
        doc (Doc): The Doc object for whose Token objects to set the attribute
            on.
        base_aspects (dict): Dictionary of keywords mapped to aspects. Should
            take the form of: {'aspect1': ['keyword1', 'keyword2'], 'aspect2':
            ['keyword3', 'keyword4']}.

    Returns:
        Doc: Processed Doc object with Token objects containing the 'aspect'
            attribute.
    """
    set_extension("aspect", target_obj=Token)

    if doc._.keywords == None:
        return doc

    for doc_keyword in doc._.keywords:
        for aspect in base_aspects:
            for keyword in base_aspects[aspect]:
                if doc_keyword.lemma_.lower() == keyword:
                    doc_keyword._.aspect = aspect
                continue
            if doc_keyword._.aspect != None:
                continue
        if doc_keyword._.aspect != None:
            continue

    return doc


def set_token_parent_span(
    doc: Doc,
    include_non_keywords: bool = False,
    min_length: int = 7,
) -> Doc:
    """Takes a Doc and adds the 'parent_span' attribute to its Token objects.

    Accessed via 'Token._.parent_span', the 'parent_span' attribute contains,
    roughly, the section of the parent Doc which pertains to a given Token. For
    a full explanation of how this is computed, see the get_parent_span()
    function.

    Target object: spacy Token
    Attribute type: Span
    Default value: None
    Dependent on: (include_non_keywords == True) N/A
        (include_non_keywords == False) set_doc_contains_aspect() ->
        set_doc_keywords() ->

    Args:
        doc (Doc): The Doc object for whose Token objects to set the attribute
            on.
        include_non_keywords (bool, optional): Whether or not to assign spans to
            non-keyword Token objects. Defaults to False.
        min_length (int, optional): Minimum span length to enforce. Spans
            shorter than the minimum length will be expanded. Set to 0 to
            disable expansion.

    Raises:
        ValueError: Raised if passing a non-bool object to include_non_keywords.

    Returns:
        Doc: Processed Doc object with Token objects containing the
            'parent_span' attribute.
    """
    set_extension("parent_span", target_obj=Token)

    if include_non_keywords == True:
        tokens = doc
    elif include_non_keywords == False and doc._.keywords is not None:
        tokens = doc._.keywords
    elif include_non_keywords == False and doc._.keywords is None:
        return doc
    else:
        raise ValueError("include_non_keywords takes only True or False")

    for token in tokens:
        span = get_token_parent_span(token, min_length=min_length)
        token._.parent_span = span

    return doc


def set_span_sentiment(
    doc: Doc,
    include_non_keywords: bool = False,  # Can only be True if parent_span True
) -> Doc:
    """Takes a Doc and adds the 'sentiment' attribute to its Span objects.

    Accessed via 'Span._.sentiment', the 'sentiment' attribute is a measure of
    the compound polarity of a span of text, as calculated by VADER (via the
    vaderSentiment package). This function calculates this sentiment for the
    parent Span objects of the Token objects in a Doc. If include_non_keywords
    is set to False, sentiment will only be calculated for the parents of
    the Doc's keywords

    Target object: spacy Span
    Attribute type: float
    Default value: None
    Dependency path: (include_non_keywords == True) set_token_parent_span() ->
        (include_non_keywords == False) set_doc_contains_aspect() ->
        set_doc_keywords() -> set_token_parent_span() ->

    Args:
        doc (Doc): The Doc object for whose Token objects to set the attribute
            on.
        include_non_keywords (bool, optional): Whether or not to assign
            sentiments to parent spans of non-keyword Token objects. Defaults to
            False.

    Raises:
        ValueError: Raised if passing a non-bool object to include_non_keywords.

    Returns:
        Doc: Processed Doc object with Span objects containing the
            'sentiment' attribute.
    """
    set_extension("sentiment", target_obj=Span)

    if include_non_keywords == True:
        tokens = doc
    elif include_non_keywords == False:
        tokens = doc._.keywords
    else:
        raise ValueError("include_non_keywords takes only True or False")

    if tokens == None:
        return doc

    for token in tokens:
        scores = ANALYZER.polarity_scores(token._.parent_span.text)
        sentiment = scores["compound"]
        token._.parent_span._.sentiment = sentiment

    return doc


def set_doc_aspect_sentiments(
    doc: Doc,
    base_aspects: dict,
) -> Doc:
    """Takes a Doc and returns a new Doc with the 'aspect_sentiments' attribute.

    Accessed via 'Doc._.aspect_sentiments', the 'aspect_sentiments' attribute
    represents the sentiment scores for all of the aspects discussed in the Doc.
    The resulting dict contains floats for all aspects that were mentioned in
    the Doc and None values for all that aren't mentioned.

    Sentiments are calculated by taking the mean of the parent span sentiments
    for all keywords mapped to each sentiment.

    Target object: spacy Doc
    Attribute type: dict
    Default value: dict (should be a dict of strings (aspects) each mapped to
        None)
    Dependency path: set_doc_contains_aspect() -> set_doc_keywords() ->
        set_token_aspects() -> set_token_parent_span() ->
        set_span_sentiment() ->

    Args:
        doc (Doc): The Doc object to set the attribute on.
        base_aspects (dict): Dictionary of keywords mapped to aspects. Should
            take the form of: {'aspect1': ['keyword1', 'keyword2'], 'aspect2':
            ['keyword3', 'keyword4']}.

    Returns:
        Doc: Processed Doc object with 'aspect_sentiments' attribute.
    """
    aspect_sentiments = {aspect: None for aspect in base_aspects}
    set_extension("aspect_sentiments", default_val=aspect_sentiments)

    if doc._.keywords is not None:
        for keyword in doc._.keywords:
            aspect = keyword._.aspect
            sentiment = keyword._.parent_span._.sentiment

            if aspect_sentiments[aspect] is None:
                aspect_sentiments[aspect] = []
            aspect_sentiments[aspect].append(sentiment)

        for aspect in aspect_sentiments:
            sentiments = aspect_sentiments[aspect]
            if sentiments == None:
                continue
            sentiment = sum(sentiments) / len(sentiments)
            aspect_sentiments[aspect] = sentiment

    doc._.aspect_sentiments = aspect_sentiments

    return doc



import spacy

from la_nlp.pipes import aspect_sentiment as asp
from la_nlp import utils

aspects = {'tests':['mid-term', 'mid term']}
keywords = utils.get_keywords_from_aspects(aspects)

nlp = spacy.load('en_core_web_lg')

rules = nlp.tokenizer.rules
rules['mid-term'] = [{65: 'mid-term'}]
rules['mid-terms'] = [{65: 'mid-terms'}]
rules['mid term'] = [{65: 'mid term'}]
nlp.tokenizer.rules = rules

text = "I wish the mid terms mid-terms had included more questions about world history."

doc = nlp(text)

doc = asp.aspect_sentiment_pipe(doc, aspects, keywords)

for kw in doc._.keywords:
    print(kw.lemma_)



"""Test functions for the la_nlp.pipes.aspect_sentiment module.
"""

import os
from la_nlp.pipes import aspect_sentiment as asp
from la_nlp import utils
import pytest
from spacy.tokens import Doc

FILE_DIR = os.path.dirname(__file__)

ASPECTS_1_PATH = os.path.join(FILE_DIR, "test_data", "test_aspects_1.toml")
ASPECTS_1 = utils.get_aspects_from_file(ASPECTS_1_PATH)

ASPECTS_2_PATH = os.path.join(FILE_DIR, "test_data", "test_aspects_2.toml")
ASPECTS_2 = utils.get_aspects_from_file(ASPECTS_2_PATH)

TEST_TEXT_1 = """
I enjoyed the course, but the readings were too long and the professor was mean.
"""

TEST_TEXT_2 = """
This is a text that does not contain any target aspects.
"""

TEST_TEXT_3 = """
The class was good. I liked the course.
"""

TEST_TEXT_4 = (
    "Professor Doe was a very passionate lecturer who presented "
    "the material quite differently from other courses I have taken. The "
    "only 'problem' I had with his course was how much bias and personal "
    "opinion they interjected in their lectures. A lot of the material "
    "presented was really just opinion and we spent too much time on that, "
    "which did not effectively facilitate learning of the subject matter. "
    "The extra material they brought in, however, was quite interesting and "
    "helped provide deeper understanding of certain subjects."
)

TEST_TEXT_5 = "The mid-terms were horrible. I wish the mid term was less boring."


@pytest.fixture
def doc1():
    doc1 = asp.make_doc(TEST_TEXT_1, aspects=ASPECTS_1, parent_span_min_length=0)
    return doc1


@pytest.fixture
def doc2():
    doc2 = asp.make_doc(TEST_TEXT_2, aspects=ASPECTS_1, parent_span_min_length=0)
    return doc2


@pytest.fixture
def doc3():
    doc3 = asp.make_doc(TEST_TEXT_3, aspects=ASPECTS_1)
    return doc3


@pytest.fixture
def doc4():
    aspects = {
        "course": ["course"],
        "format": ["format"],
        "content": ["content"],
        "professor": ["professor"],
    }
    doc4 = asp.make_doc(TEST_TEXT_4, aspects=aspects)
    return doc4


@pytest.fixture
def doc5():
    aspects = {"tests": ["mid-term", "mid term"]}
    doc5 = asp.make_doc(TEST_TEXT_5, aspects=aspects)
    return doc5


def test_function_make_doc(doc1, doc3):
    """Tests that the make_doc() function returns a spacy Doc object."""
    assertion1 = "Should be a spacy Doc object"
    assert isinstance(doc1, Doc), assertion1

    assertion2 = "Should have a length of 19"
    assert len(doc1) == 19, assertion2

    assert isinstance(doc3, Doc)


def test_attribute_contains_aspect(doc1, doc2):
    """Tests that docs are assigned the contains_aspect attribute as expected."""
    assertion1 = "doc1 should return True"
    assert doc1._.contains_aspect == True, assertion1

    assertion2 = "doc2 should return False"
    assert doc2._.contains_aspect == False, assertion2


def test_attribute_aspects(doc1, doc2):
    """Tests that docs are assigned the aspects attribute as expected."""
    assertion1 = "doc1 should contain 'course', 'content', and 'instructor'"
    doc1_aspects = ["course", "content", "instructor"]
    assert doc1._.aspects == doc1_aspects, assertion1

    assertion2 = "doc2 should contain None"
    doc2_aspects = None
    assert doc2._.aspects == doc2_aspects, assertion2


def test_attribute_keywords(doc1, doc2):
    """Tests that docs are assigned the keywords attribute as expected."""
    assertion1 = "doc1 should contain keyword %s"
    doc1_targets = ["course", "reading", "professor"]
    doc1_keywords = [kw.lemma_.lower() for kw in doc1._.keywords]
    for target in doc1_targets:
        assert target in doc1_keywords, assertion1 % target

    assertion2 = "doc2 should contain no keywords"
    doc2_targets = None
    assert doc2._.keywords == doc2_targets, assertion2


def test_attribute_keyword_aspects(doc1):
    """Tests that tokens are assigned the aspect attribute as expected."""
    assertion = "%s aspect should be %s"
    target_aspects = ["course", "content", "instructor"]
    for keyword, aspect in zip(doc1._.keywords, target_aspects):
        assert keyword._.aspect == aspect, assertion % (keyword.text, aspect)


def test_attribute_parent_span(doc1, doc2):
    """Tests that tokens are assigned the parent span attribute as expected."""
    assertion1 = "Span should read 'the professor was mean'"
    doc1_target = "the professor was mean"
    token1 = doc1._.keywords[2]
    assert token1._.parent_span.text == doc1_target, assertion1

    assertion2 = "Span should be None"
    doc2_target = None
    token2 = doc2[-3]
    assert token2._.parent_span == doc2_target, assertion2


def test_attribute_parent_span_length(doc4):
    """Tests that token parent spans are above minimum length."""
    kw = doc4._.keywords[0]
    span = kw._.parent_span
    target = (
        "Professor Doe was a very passionate lecturer who presented the "
        "material quite differently from other courses I have taken."
    )
    assertion = f"Span should read {target}"

    assert span.text == target, assertion


def test_attribute_parent_span_sentiment(doc1):
    """Tests that spans are assigned sentiment attribute as expected."""
    assertion = "course parent span sentiment should be 0.2846"
    span = doc1._.keywords[0]._.parent_span
    target_sentiment = 0.2846

    assert span._.sentiment == target_sentiment, assertion


def test_attribute_aspect_sentiments(doc1, doc3):
    """Tests that docs are assigned aspect sentiments attribute as expected."""
    target_sentiments1 = {
        "course": 0.2846,
        "content": 0,
        "assignments": None,
        "tests": None,
        "instructor": 0,
    }
    assertion1 = f"Sentiment dict should look like: {target_sentiments1}"

    assert doc1._.aspect_sentiments == target_sentiments1, assertion1

    target_course_sentiment = 0.43095
    course_sentiment = doc3._.aspect_sentiments["course"]

    assertion2 = "doc3 course sentiment value should within 0.00001 of 0.43095"

    assert abs(target_course_sentiment - course_sentiment) <= 0.00001, assertion2


def test_make_doc_with_aspects_from_file():
    """Tests that make_doc() function is able to accept aspects from file path."""
    assertion = "Should successfully take aspects from a file path"
    aspects = ASPECTS_1_PATH
    doc = asp.make_doc(TEST_TEXT_1, aspects=aspects)

    assert doc._.aspect_sentiments is not None, assertion


def test_make_doc_error_from_non_path_aspects_string():
    """Tests that make_doc() function returns ValueError for non-path strings."""
    assertion = "Should raise a ValueError for passing a non-path string to aspects"
    aspects = "This is a string that should raise an error"
    with pytest.raises(ValueError) as excinfo:
        target_error = "Aspects must be either a dict or path to .toml file"
        asp.make_doc(TEST_TEXT_1, aspects=aspects)
        assert excinfo.value == target_error, assertion


def test_multi_word_expression_keywords(doc5):
    """Tests that multi-word expressions in aspect keywords are not split by tokenizer"""
    target_len = 13
    assertion1 = f"Length of doc should be {target_len} tokens"
    assert len(doc5) == target_len, assertion1

    target_kws = ["mid-term", "mid term"]
    kws = [kw.lemma_ for kw in doc5._.keywords]
    assertion2 = f"Keywords in doc should {target_kws}"
    assert kws == target_kws, assertion2



"""Test functions for the la_nlp.utils module.
"""

import os
from la_nlp import utils

FILE_DIR = os.path.dirname(__file__)

ASPECTS_1 = os.path.join(FILE_DIR, "test_data", "test_aspects_1.toml")
ASPECTS_2 = os.path.join(FILE_DIR, "test_data", "test_aspects_2.toml")


def test_get_aspects_from_file():
    """Tests that get_aspects_from_file() is retrieving aspects as expected."""
    aspects = utils.get_aspects_from_file(ASPECTS_1)
    target = {
        "course": ["course", "class", "lecture"],
        "content": ["content", "material", "reading", "syllabus", "powerpoint"],
        "assignments": ["assignment", "project", "paper", "homework"],
        "tests": [
            "test",
            "quiz",
            "exam",
            "examination",
            "midterm",
            "mid-term",
            "final",
        ],
        "instructor": ["instructor", "teacher", "professor", "prof", "dr"],
    }
    assertion = f"Aspects should be {aspects}."
    assert aspects == target, assertion


def test_keywords_from_aspects():
    """Tests that get_keywords_from_aspects() is retrieving keywords properly."""
    aspects = utils.get_aspects_from_file(ASPECTS_1)
    keywords = utils.get_keywords_from_aspects(aspects)
    target = [
        "course",
        "class",
        "lecture",
        "content",
        "material",
        "reading",
        "syllabus",
        "powerpoint",
        "assignment",
        "project",
        "paper",
        "homework",
        "test",
        "quiz",
        "exam",
        "examination",
        "midterm",
        "mid-term",
        "final",
        "instructor",
        "teacher",
        "professor",
        "prof",
        "dr",
    ]
    assertion = f"Keywords should be {target}."
    assert keywords == target, assertion



"""Utilities for use in other modules within package.
"""

import os
import tomllib

FILE_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.dirname(FILE_DIR)
DATA_DIR = os.path.join(FILE_DIR, "data")
ASPECT_FILE = os.path.join(DATA_DIR, "aspects.toml")


def get_aspects_from_file(
    file_path: str,
) -> dict:
    """Gets dictionary of aspects from .toml aspects file.

    Args:
        file_path (str): Path to aspects file. Must be a .toml file in the shape
            described in documentation.

    Returns:
        dict: Dictionary of aspects and corresponding keywords.
    """
    with open(file_path, "rb") as file:
        aspects = tomllib.load(file)
    return aspects


def get_keywords_from_aspects(
    aspects: dict,
) -> list:
    """Extracts list of keywords from dictionary of aspects.

    Args:
        aspects (dict): Dictionary of aspects and corresponding keywords.

    Returns:
        list: List of all keywords contained in aspects dict.
    """
    keywords = []
    for keywords_list in aspects.values():
        keywords.extend(keywords_list)
    return keywords


def get_default_aspects() -> dict:
    """Retrieves dictionary of default aspects in la_nlp/data/aspects.toml.

    Returns:
        dict: Dictionary of aspects and corresponding keywords.
    """
    return get_aspects_from_file(ASPECT_FILE)



"""NLP pipeline for performing aspect-based sentiment analysis with spacy.

This module contains a pipeline for performing aspect-based sentiment analysis
via the spacy NLP package and the VADER sentiment model. The pipeline can be
called via the make_doc() function, which accepts a string as input and returns
a spacy Doc object containing a number of attributes useful for this type of
analysis. See documentation for a list of attributes assigned by this pipeline.
"""

import os
import re

from la_nlp import components, utils

from spacy import load as load_model
from spacy.language import Language
from spacy.tokens import Doc

DEFAULT_ASPECTS = utils.get_default_aspects()

NLP = load_model("en_core_web_lg")


def make_doc(
    text: str,
    aspects: dict | str = DEFAULT_ASPECTS,
    parent_span_min_length: int = 7,
) -> Doc:
    """Generates a spacy Doc object via the aspect sentiment pipeline.

    Args:
        text (str): The text to process.
        aspects (dict | str, optional): The aspects to use for aspect-based
            sentiment analysis. Can be either a dictionary of aspects with
            corresponding arrays of keywords, or a path to a .toml file
            containing the aspect-keyword mappings. Defaults to default aspects
            at la_nlp/data/aspects.toml.
        parent_span_min_length (int, optional): Minimum length from which to
            generate token parent spans. Defaults to 7.

    Raises:
        ValueError: Raised if value passed to aspects is not a file path or
            a dictionary.

    Returns:
        Doc: Processed Doc object from input text containing attributes
            generated by the aspect_sentiment pipeline.
    """

    def except_multi_word_expressions(keywords: list) -> None:
        """Creates exception for multi-word expression keywords to not be tokenized.

        Checks for keywords containing characters used for token splitting. For each
        keyword containing a splitter, adds the keyword and its pluralized form to the
        spacy tokenizer as an exception.
        """
        rules = NLP.tokenizer.rules
        regex = r"[-\s/']"
        for keyword in keywords:
            if keyword in rules:
                continue
            if re.search(regex, keyword):
                rules[keyword] = [{65: keyword}]
                rules[keyword + "s"] = [{65: keyword + "s"}]
        NLP.tokenizer.rules = rules

    if isinstance(aspects, str) and os.path.isfile(aspects):
        aspects = utils.get_aspects_from_file(aspects)
    elif not isinstance(aspects, dict):
        raise ValueError("Aspects must be either a dict or path to .toml file")

    keywords = utils.get_keywords_from_aspects(aspects)
    except_multi_word_expressions(keywords)

    cfg = {
        "aspect_sentiment_pipe": {
            "aspects": aspects,
            "keywords": keywords,
            "parent_span_min_length": parent_span_min_length,
        }
    }
    return NLP(text, component_cfg=cfg)


@Language.component("aspect_sentiment_pipe")
def aspect_sentiment_pipe(
    doc: Doc,
    aspects: dict,
    keywords: list,
    parent_span_min_length: int = 7,
) -> Doc:
    """Compiles the pipeline components into a single function.

    Should not be called publically. This function is only used as an interface
    between the spacy pipeline and the pipeline components contained within this
    package.
    """
    doc = components.set_doc_contains_aspect(doc, keywords)
    doc = components.set_doc_aspects(doc, aspects)
    doc = components.set_doc_keywords(doc, keywords)
    doc = components.set_token_aspects(doc, aspects)
    doc = components.set_token_parent_span(doc, min_length=parent_span_min_length)
    doc = components.set_span_sentiment(doc)
    doc = components.set_doc_aspect_sentiments(doc, aspects)
    return doc


NLP.add_pipe("aspect_sentiment_pipe")



"""Test functions for the la_nlp.components module.
"""

from la_nlp import components as comp
from spacy import load as load_model
from spacy.tokens import Doc
import pytest

TEST_TEXT_1 = """
This is a text that contains keyword1.
"""

ASPECTS = {"aspect1": ["keyword1"]}
KEYWORDS = ["keyword1"]


@pytest.fixture
def nlp():
    nlp = load_model("en_core_web_lg")
    return nlp


def test_function_get_token_parent_span(nlp):
    text = "Hello world!"
    doc = nlp(text)
    assert comp.get_token_parent_span(doc[0], min_length=7).text == text


def test_function_contains_aspect(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    assert doc._.contains_aspect is not None


def test_function_aspects(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    doc = comp.set_doc_aspects(doc, base_aspects=ASPECTS)
    assert doc._.aspects is not None


def test_function_keywords(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    doc = comp.set_doc_keywords(doc, base_keywords=KEYWORDS)
    assert doc._.keywords is not None


def test_function_keyword_aspects(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    doc = comp.set_doc_keywords(doc, base_keywords=KEYWORDS)
    doc = comp.set_token_aspects(doc, base_aspects=ASPECTS)
    assert doc._.keywords[0]._.aspect is not None


def test_function_parent_span(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    doc = comp.set_doc_keywords(doc, base_keywords=KEYWORDS)
    doc = comp.set_token_parent_span(doc)
    token1 = doc._.keywords[0]
    token2 = doc[1]
    assert token1._.parent_span is not None
    assert token2._.parent_span is None


def test_function_parent_span_non_keywords(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_token_parent_span(doc, include_non_keywords=True)
    assert doc[0]._.parent_span is not None


def test_function_parent_span_sentiment(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    doc = comp.set_doc_keywords(doc, base_keywords=KEYWORDS)
    doc = comp.set_token_parent_span(doc)
    doc = comp.set_span_sentiment(doc)
    token = doc._.keywords[0]
    assert token._.parent_span._.sentiment is not None


def test_function_parent_span_sentiment_non_keywords(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_token_parent_span(doc, include_non_keywords=True)
    doc = comp.set_span_sentiment(doc, include_non_keywords=True)
    assert doc[0]._.parent_span._.sentiment is not None


def test_function_aspect_sentiments(nlp):
    doc = nlp(TEST_TEXT_1)
    doc = comp.set_doc_contains_aspect(doc, base_keywords=KEYWORDS)
    doc = comp.set_doc_keywords(doc, base_keywords=KEYWORDS)
    doc = comp.set_token_aspects(doc, base_aspects=ASPECTS)
    doc = comp.set_token_parent_span(doc)
    doc = comp.set_span_sentiment(doc)
    doc = comp.set_doc_aspect_sentiments(doc, base_aspects=ASPECTS)
    assert doc._.aspect_sentiments is not None
